"""
The gtfs_segments package main init file.
"""
__version__ = "0.2.0"
from .geom_utils import view_heatmap, view_spacings, view_spacings_interactive
from .gtfs_segments import get_gtfs_segments, pipeline_gtfs, process_feed
from .mobility import (
    download_latest_data,
    fetch_gtfs_source,
    summary_stats_mobility,
)
from .partridge_func import get_bus_feed
from .route_stats import get_route_stats
from .utils import export_segments, plot_hist, process, summary_stats

__all__ = [
    "__version__",
    "get_gtfs_segments",
    "pipeline_gtfs",
    "process_feed",
    "export_segments",
    "plot_hist",
    "fetch_gtfs_source",
    "summary_stats",
    "process",
    "view_spacings",
    "view_spacings_interactive",
    "view_heatmap",
    "summary_stats_mobility",
    "download_latest_data",
    "get_route_stats",
    "get_bus_feed",
]



import os
import unittest

import pandas as pd

from gtfs_segments import get_bus_feed, get_route_stats

test_dir = os.path.dirname(__file__)


class TestRouteStats(unittest.TestCase):
    """Tests for utils.py module in the package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.gtfs_path = os.path.join(
            test_dir,
            "data",
            "Ann Arbor-University of Michigan Transit Services-MI",
            "gtfs.zip",
        )

    def test_route_stats(self):
        feed = get_bus_feed(self.gtfs_path)
        route_stats_df = get_route_stats(feed)
        self.assertTrue(
            isinstance(route_stats_df, pd.DataFrame),
            "Error with summary_stats. Result should be a DataFrame",
        )



import os
from typing import Optional

import pandas as pd

import gtfs_segments.partridge_mod as ptg

from .partridge_mod.gtfs import Feed, parallel_read


def get_bus_feed(
    path: str, agency_id: Optional[str] = None, threshold: Optional[int] = 1, parallel: bool = False
) -> Feed:
    """
    The `get_bus_feed` function retrieves bus feed data from a specified path, with the option to filter
    by agency name, and returns the busiest date and a GTFS feed object.

    Args:
      path (str): The `path` parameter is a string that represents the path to the GTFS file. This file
    contains the bus feed data.
      agency_id (All): The `agency_id` parameter is an optional parameter that allows you to filter the
    bus feed data by the agency name. It is used to specify the ID of the transit agency for which you
    want to retrieve the bus feed data. If you provide an `agency_id`, the function will only return
    data
      threshold (int): The `threshold` parameter is used to filter out service IDs that have a low
    frequency. It is set to a default value of 1, but you can adjust it to a different value if needed.
    Service IDs with a sum of stop times greater than the threshold will be included in the returned
    bus. Defaults to 1

    Returns:
      A tuple containing the busiest date and a GTFS feed object. The GTFS feed object contains
    information about routes, stops, stop times, trips, and shapes for a transit agency's schedule.
    """
    b_day, bday_service_ids = ptg.read_busiest_date(path)
    print("Using the busiest day:", b_day)
    all_days_s_ids_df = get_all_days_s_ids(path)
    series = all_days_s_ids_df[bday_service_ids].sum(axis=0) > threshold
    service_ids = series[series].index.values
    route_types = [3, 700, 702, 703, 704, 705]  # 701 is regional
    # set of service IDs eliminated due to low frequency
    removed_service_ids = set(bday_service_ids) - set(service_ids)
    if len(removed_service_ids) > 0:
        print("Service IDs eliminated due to low frequency:", removed_service_ids)
    if agency_id is not None:
        view = {
            "routes.txt": {"route_type": route_types},  # Only bus routes
            "trips.txt": {"service_id": service_ids},  # Busiest day only
            "agency.txt": {"agency_id": agency_id},  # Eg: 'Société de transport de Montréal
        }
    else:
        view = {
            "routes.txt": {"route_type": route_types},  # Only bus routes
            "trips.txt": {"service_id": service_ids},  # Busiest day only
        }
    feed = ptg.load_geo_feed(path, view=view)
    if parallel:
        num_cores = os.cpu_count()
        print(":: Processing Feed in Parallel :: Number of cores:", num_cores)
        parallel_read(feed)
    return feed


def get_all_days_s_ids(path: str) -> pd.DataFrame:
    """
    Read dates by service IDs from a given path, create a DataFrame, populate it with the dates and
    service IDs, and fill missing values with False.

    Args:
      path: The path to the GTFS file

    Returns:
      A DataFrame containing dates and service IDs.
    """
    dates_by_service_ids = ptg.read_dates_by_service_ids(path)
    data = dates_by_service_ids
    # Create a DataFrame
    data_frame = pd.DataFrame(columns=sorted(list({col for row in data.keys() for col in row})))

    # Iterate through the data and populate the DataFrame
    for service_ids, dates in data.items():
        for date_value in dates:
            data_frame.loc[date_value, list(service_ids)] = True

    # Fill missing values with False
    data_frame.fillna(False, inplace=True)
    return data_frame



import datetime
import os
import unittest

import geopandas as gpd

import gtfs_segments.partridge_mod as ptg

test_dir = os.path.dirname(__file__)


class TestPartridge(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures, if any."""
        self.gtfs_path = os.path.join(
            test_dir,
            "data",
            "Ann Arbor-University of Michigan Transit Services-MI",
            "gtfs.zip",
        )

    def test_load_geo_feed(self):
        feed = ptg.load_geo_feed(self.gtfs_path)
        assert isinstance(feed.shapes, gpd.GeoDataFrame)
        assert isinstance(feed.stops, gpd.GeoDataFrame)
        assert {"LineString"} == set(feed.shapes.geom_type)
        assert {"Point"} == set(feed.stops.geom_type)
        assert feed.shapes.crs == "epsg:4326"
        assert feed.stops.crs == "epsg:4326"
        assert ["shape_id", "geometry"] == list(feed.shapes.columns)
        # assert [
        #     "stop_id",
        #     "stop_code",
        #     "stop_name",
        #     "stop_desc",
        #     "zone_id",
        #     "stop_url",
        #     "location_type",
        #     "parent_station",
        #     "stop_timezone",
        #     "wheelchair_boarding",
        #     "geometry",
        # ] == list(feed.stops.columns)

    def test_service_ids_by_date(self):
        service_ids_by_date = ptg.read_service_ids_by_date(self.gtfs_path)

        assert len(service_ids_by_date) == 133
        assert datetime.date(2022, 2, 17) in service_ids_by_date

    def test_dates_by_service_ids(self):
        dates_by_service_ids = ptg.read_dates_by_service_ids(self.gtfs_path)
        assert len(dates_by_service_ids) == 16
        assert frozenset({"2"}) in dates_by_service_ids

    def test_trip_counts_by_date(self):
        trip_counts_by_date = ptg.read_trip_counts_by_date(self.gtfs_path)
        assert len(trip_counts_by_date) == 133
        assert datetime.date(2022, 4, 12) in trip_counts_by_date

    def test_busiest_date(self):
        date, service_ids = ptg.read_busiest_date(self.gtfs_path)
        assert date == datetime.date(2022, 1, 10)
        assert service_ids == frozenset({"7"})

    def test_busiest_week(self):
        service_ids_by_date = ptg.read_busiest_week(self.gtfs_path)
        assert service_ids_by_date == {
            datetime.date(2022, 1, 10): frozenset({"7"}),
            datetime.date(2022, 1, 11): frozenset({"10"}),
            datetime.date(2022, 1, 12): frozenset({"10"}),
            datetime.date(2022, 1, 13): frozenset({"10"}),
            datetime.date(2022, 1, 14): frozenset({"4"}),
            datetime.date(2022, 1, 15): frozenset({"11"}),
            datetime.date(2022, 1, 16): frozenset({"9"}),
        }



"""Tests for `gtfs_segments` package."""

import os
import unittest

import geopandas as gpd
import matplotlib
import pandas as pd

from gtfs_segments import export_segments, get_gtfs_segments, plot_hist, summary_stats

test_dir = os.path.dirname(__file__)
test_folder = os.path.join(test_dir, "output")
if not os.path.exists(test_folder):
    os.mkdir(test_folder)


class TestUtils(unittest.TestCase):
    """Tests for utils.py module in the package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        # Using UMich Ann Arbor as a test feed to save time
        self.gtfs_path = os.path.join(
            test_dir,
            "data",
            "Ann Arbor-University of Michigan Transit Services-MI",
            "gtfs.zip",
        )
        self.df = get_gtfs_segments(self.gtfs_path)

    def test_plot_histogram(self):
        fig = plot_hist(self.df)
        self.assertTrue(
            type(fig) == matplotlib.figure.Figure,
            "Error with plot_hist. Should work for the SFMTA example feed",
        )
        fig = plot_hist(self.df, show_mean=True)
        self.assertTrue(
            type(fig) == matplotlib.figure.Figure,
            "Error with plot_hist. Should work for the SFMTA example feed",
        )
        fig = plot_hist(
            self.df,
            show_mean=True,
            save_fig=True,
            file_path=os.path.join(test_dir, "output", "test_hist.png"),
        )
        self.assertTrue(
            type(fig) == matplotlib.figure.Figure,
            "Error with plot_hist. Should work for the SFMTA example feed",
        )
        self.assertTrue(
            os.path.exists(os.path.join(test_dir, "output", "test_hist.png")),
            "Check if the test_hist.png file exists",
        )

    def test_summary_stats(self):
        summ_df = summary_stats(self.df)
        self.assertTrue(
            type(summ_df) == pd.DataFrame,
            "Error with summary_stats. Result should be a DataFrame",
        )
        summ_df = summary_stats(
            self.df,
            export=True,
            file_path=os.path.join(test_dir, "output", "test_summary_stats.csv"),
        )
        self.assertTrue(
            isinstance(summ_df, pd.DataFrame),
            "Error with summary_stats. Result should be a DataFrame",
        )
        self.assertTrue(
            os.path.exists(os.path.join(test_dir, "output", "test_summary_stats.csv")),
            "Check if the test_summary_stats.csv file exists",
        )
        self.assertTrue(
            sum(summ_df[summ_df.columns[0]] < 0) == 0,
            "Summary stats should be positive",
        )

    def test_export_segments(self):
        # Test export_segments for .csv and with geometry
        export_segments(
            self.df,
            os.path.join(test_dir, "output", "test_export_segments.csv"),
            output_format="csv",
        )
        self.assertTrue(
            os.path.exists(os.path.join(test_dir, "output", "test_export_segments.csv")),
            "Check if the test_export_segments.csv file exists",
        )
        df = pd.read_csv(os.path.join(test_dir, "output", "test_export_segments.csv"))
        self.assertTrue(
            type(df) == pd.DataFrame,
            "Import again to see if the file is valid and has the same data",
        )
        self.assertTrue(
            len(df) == len(self.df),
            "Error with export_segments. Should work for the SFMTA example feed",
        )

        # Test export_segments for .csv and without geometry
        export_segments(
            self.df,
            os.path.join(test_dir, "output", "test_export_segments.csv"),
            output_format="csv",
            geometry=False,
        )
        self.assertTrue(
            os.path.exists(os.path.join(test_dir, "output", "test_export_segments.csv")),
            "Check if the test_export_segments.csv file exists",
        )
        df = pd.read_csv(os.path.join(test_dir, "output", "test_export_segments.csv"))
        self.assertTrue(
            type(df) == pd.DataFrame,
            "Import again to see if the file is valid and has the same data",
        )
        self.assertTrue("geometry" not in df.columns, "Check if the geometry column is not present")

        # Test export_segments for .geojson and with geometry
        export_segments(
            self.df,
            os.path.join(test_dir, "output", "test_export_segments.geojson"),
            output_format="geojson",
        )
        self.assertTrue(
            os.path.exists(os.path.join(test_dir, "output", "test_export_segments.geojson")),
            "Check if the test_export_segments.geojson file exists",
        )
        gdf = gpd.read_file(os.path.join(test_dir, "output", "test_export_segments.geojson"))
        self.assertTrue(
            type(gdf) == gpd.GeoDataFrame,
            "Import again to see if the file is valid and has the same data",
        )
        self.assertTrue(
            len(gdf) == len(self.df),
            "Error with export_segments. Should work for the SFMTA example feed",
        )



from typing import Any, BinaryIO, Set

from charset_normalizer import detect
from pandas.core.common import flatten


def setwrap(value: Any) -> Set[str]:
    """
    Returns a flattened and stringified set from the given object or iterable.

    For use in public functions which accept argmuents or kwargs that can be
    one object or a list of objects.
    """
    return set(map(str, set(flatten([value]))))


def detect_encoding(file: BinaryIO, limit: int = 2500) -> str:
    """
    Return encoding of provided input stream.

    Most of the time it's unicode, but if we are unable to decode the input
    natively, use `chardet` to determine the encoding heuristically.
    """
    unicode_decodable = True

    for line_no, line in enumerate(file):
        try:
            line.decode("utf-8")
        except UnicodeDecodeError:
            unicode_decodable = False
            break

        if line_no > limit:
            break

    if unicode_decodable:
        return "utf-8"

    file.seek(0)
    encoding = detect(file.read())["encoding"]
    return encoding



import datetime
import os
import shutil
import tempfile
import weakref
from collections import defaultdict
from typing import DefaultDict, Dict, FrozenSet, Optional, Set, Tuple

from isoweek import Week

# from .config import default_config, geo_config, empty_config, reroot_graph
from .gtfs import Feed, View
from .parsers import vparse_date

DAY_NAMES = (
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
)


def load_feed(path: str, view: Optional[View] = None) -> Feed:
    # config = default_config() if config is None else config
    view = {} if view is None else view

    # if not nx.is_directed_acyclic_graph(config):
    #     raise ValueError("Config must be a DAG")

    if os.path.isdir(path):
        feed = _load_feed(path, view)
    elif os.path.isfile(path):
        feed = _unpack_feed(path, view)
    else:
        raise ValueError("File or path not found: {}".format(path))

    return feed


def load_raw_feed(path: str) -> Feed:
    return load_feed(path, view={})


def load_geo_feed(path: str, view: Optional[View] = None) -> Feed:
    return load_feed(path, view=view)


def read_busiest_date_feed(path: str) -> Tuple[Feed, datetime.date, FrozenSet[str]]:
    """Find the earliest date with the most trips"""
    feed = load_raw_feed(path)
    date, service_ids = _busiest_date(feed)
    return feed, date, service_ids


def read_busiest_date(path: str) -> Tuple[datetime.date, FrozenSet[str]]:
    """Find the earliest date with the most trips"""
    feed = load_raw_feed(path)
    return _busiest_date(feed)


def read_busiest_week(path: str) -> Dict[datetime.date, FrozenSet[str]]:
    """Find the earliest week with the most trips"""
    feed = load_raw_feed(path)
    return _busiest_week(feed)


def read_service_ids_by_date(path: str) -> Dict[datetime.date, FrozenSet[str]]:
    """Find all service identifiers by date"""
    feed = load_raw_feed(path)
    return _service_ids_by_date(feed)


def read_dates_by_service_ids(
    path: str,
) -> Dict[FrozenSet[str], FrozenSet[datetime.date]]:
    """Find dates with identical service"""
    feed = load_raw_feed(path)
    return _dates_by_service_ids(feed)


def read_trip_counts_by_date(path: str) -> Dict[datetime.date, int]:
    """A useful proxy for busyness"""
    feed = load_raw_feed(path)
    return _trip_counts_by_date(feed)


def _unpack_feed(path: str, view: View) -> Feed:
    tmpdir = tempfile.mkdtemp()
    shutil.unpack_archive(path, tmpdir)
    feed: Feed = _load_feed(tmpdir, view)

    # Eager cleanup
    feed._delete_after_reading = True

    def finalize() -> None:
        shutil.rmtree(tmpdir)

    # Lazy cleanup
    weakref.finalize(feed, finalize)

    return feed


def _load_feed(path: str, view: View) -> Feed:
    """Multi-file feed filtering"""
    feed_ = Feed(path, view={})
    for filename, column_filters in view.items():
        view_ = {filename: column_filters}
        feed_ = Feed(feed_, view=view_)
    return Feed(feed_)


def _busiest_date(feed: Feed) -> Tuple[datetime.date, FrozenSet[str]]:
    service_ids_by_date = _service_ids_by_date(feed)
    trip_counts_by_date = _trip_counts_by_date(feed)

    def max_by(kv: Tuple[datetime.date, int]) -> Tuple[int, int]:
        date, count = kv
        return count, -date.toordinal()

    date, _ = max(trip_counts_by_date.items(), key=max_by)
    service_ids = service_ids_by_date[date]

    return date, service_ids


def _busiest_week(feed: Feed) -> Dict[datetime.date, FrozenSet[str]]:
    service_ids_by_date = _service_ids_by_date(feed)
    trip_counts_by_date = _trip_counts_by_date(feed)

    weekly_trip_counts: DefaultDict[Week, int] = defaultdict(int)
    weekly_dates: DefaultDict[Week, Set[datetime.date]] = defaultdict(set)
    for date in service_ids_by_date.keys():
        key = Week.withdate(date)
        weekly_trip_counts[key] += trip_counts_by_date[date]
        weekly_dates[key].add(date)

    def max_by(kv: Tuple[Week, int]) -> Tuple[int, int]:
        week, count = kv
        return count, -week.toordinal()

    week, _ = max(weekly_trip_counts.items(), key=max_by)
    dates = sorted(weekly_dates[week])

    return {date: service_ids_by_date[date] for date in dates}


def _service_ids_by_date(feed: Feed) -> Dict[datetime.date, FrozenSet[str]]:
    results: DefaultDict[datetime.date, Set[str]] = defaultdict(set)
    removals: DefaultDict[datetime.date, Set[str]] = defaultdict(set)

    service_ids = set(feed.trips.service_id)
    calendar = feed.calendar
    caldates = feed.calendar_dates

    if not calendar.empty:
        # Only consider calendar.txt rows with applicable trips
        calendar = calendar[calendar.service_id.isin(service_ids)].copy()

    if not caldates.empty:
        # Only consider calendar_dates.txt rows with applicable trips
        caldates = caldates[caldates.service_id.isin(service_ids)].copy()

    if not calendar.empty:
        # Parse dates
        calendar.start_date = vparse_date(calendar.start_date)
        calendar.end_date = vparse_date(calendar.end_date)

        # Build up results dict from calendar ranges
        for _, cal in calendar.iterrows():
            start = cal.start_date.toordinal()
            end = cal.end_date.toordinal()

            dow = {i: cal[day] for i, day in enumerate(DAY_NAMES)}
            for ordinal in range(start, end + 1):
                date = datetime.date.fromordinal(ordinal)
                if int(dow[date.weekday()]):
                    results[date].add(cal.service_id)

    if not caldates.empty:
        # Parse dates
        caldates.date = vparse_date(caldates.date)

        # Split out additions and removals
        cdadd = caldates[caldates.exception_type == 1]
        cdrem = caldates[caldates.exception_type == 2]

        # Add to results by date
        for _, cd in cdadd.iterrows():
            results[cd.date].add(cd.service_id)

        # Collect removals
        for _, cd in cdrem.iterrows():
            removals[cd.date].add(cd.service_id)

        # Finally, process removals by date
        for date in removals:
            for service_id in removals[date]:
                if service_id in results[date]:
                    results[date].remove(service_id)

            # Drop the key from results if no service present
            if len(results[date]) == 0:
                del results[date]

    assert results, "No service found in feed."

    return {k: frozenset(v) for k, v in results.items()}


def _dates_by_service_ids(feed: Feed) -> Dict[FrozenSet[str], FrozenSet[datetime.date]]:
    results: DefaultDict[FrozenSet[str], Set[datetime.date]] = defaultdict(set)
    for date, service_ids in _service_ids_by_date(feed).items():
        results[service_ids].add(date)
    return {k: frozenset(v) for k, v in results.items()}


def _trip_counts_by_date(feed: Feed) -> Dict[datetime.date, int]:
    results: DefaultDict[datetime.date, int] = defaultdict(int)
    trips = feed.trips
    for service_ids, dates in _dates_by_service_ids(feed).items():
        trip_count = trips[trips.service_id.isin(service_ids)].shape[0]
        for date in dates:
            results[date] += trip_count
    return dict(results)



from concurrent.futures import ThreadPoolExecutor
from typing import Any, List, Optional, Tuple

import contextily as cx
import folium
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import utm
from matplotlib.figure import Figure
from pyproj import Geod
from scipy.spatial import cKDTree
from shapely.geometry import LineString, Point
from shapely.ops import split

geod = Geod(ellps="WGS84")


def split_route(row: pd.Series) -> str:
    """
    It takes a row from a dataframe, and if the row has a start and end point,
    it splits the route into two segments

    Args:
      row: row in stop_df

    Returns:
      the geometry of the route segments.
    """
    route = row["geometry"]
    if row["snapped_start_id"]:
        try:
            route = split(route, row["start"]).geoms[1]
        except IndexError:
            pass
    if row["snapped_end_id"]:
        route = split(route, row["end"]).geoms[0]
    return route.wkt


def nearest_snap(route_string: LineString, stop_point: Point) -> str:
    """
    It takes a dataframe of bus stops and a dataframe of bus routes and
    returns a dataframe of the nearest bus stop to each bus stop

    Args:
      route: the route geometry
      point: the point you want to snap to the nearest point on the route

    Returns:
      A list of tuples. Each tuple contains the route_id, segment_id, and
      the distance between the two stops.
    """
    route = np.array(route_string.coords)
    point = np.array(stop_point.coords)
    ckd_tree = cKDTree(route)
    return Point(route[ckd_tree.query(point, k=1)[1]][0]).wkt


def make_gdf(df: pd.DataFrame) -> gpd.GeoDataFrame:
    """
    It takes a dataframe and returns a geodataframe

    Args:
      df: the dataframe you want to convert to a geodataframe

    Returns:
      A GeoDataFrame
    """
    gdf = gpd.GeoDataFrame(df)
    gdf = gdf.set_crs(epsg=4326, allow_override=True)
    return gdf


def code(zone: List, lat: float) -> int:
    """
    If the latitude is negative, the EPSG code is 32700 + the zone number. If
    the latitude is positive, the EPSG code is 32600 + the zone number

    Args:
      zone: The UTM zone number.
      lat: latitude of the point

    Returns:
      The EPSG Code
    """
    if lat < 0:
        epsg_code = 32700 + zone[2]
    else:
        epsg_code = 32600 + zone[2]
    return epsg_code


def get_zone_epsg(stop_df: gpd.GeoDataFrame) -> int:
    """
    > The function takes a dataframe with a geometry column and returns the
    EPSG code for the UTM zone that contains the geometry

    Args:
      stop_df: a dataframe with a geometry column

    Returns:
      The EPSG code for the UTM zone that the stop is in.
    """
    lon = stop_df.start[0].x
    lat = stop_df.start[0].y
    zone = utm.from_latlon(lat, lon)
    return code(zone, lat)


def view_spacings(
    gdf: gpd.GeoDataFrame,
    basemap: bool = False,
    map_provider: str = cx.providers.CartoDB.Positron,
    show_stops: bool = False,
    level: str = "whole",
    axis: str = "on",
    dpi: int = 300,
    **kwargs: Any,
) -> Figure:
    """
    The `view_spacings` function plots the spacings of a bus network, route, or segment, with options to
    add a basemap and show stops.

    Args:
      gdf: The GTFS segments GeoDataframe containing the bus network data.
      basemap: The `basemap` parameter is a boolean value that determines whether to add a basemap to
    the plot. If set to `True`, a basemap will be added. If set to `False`, no basemap will be added.
    The default value is `False`. Defaults to False
      map_provider: The `map_provider` parameter is used to specify the source of the basemap that
    will be added to the plot. It is set to `cx.providers.CartoDB.Positron` by default, which means
    that the basemap will be sourced from CartoDB's Positron. Use `contextily.providers` to see full
    list of providers
      show_stops: The `show_stops` parameter is a boolean flag that determines whether or not to display
    the bus stops on the plot. If set to `True`, the bus stops will be shown as white markers on the
    plot. If set to `False`, the bus stops will not be shown. Defaults to False
      level: The "level" parameter determines the level of detail to plot the spacings. It can take one
    of three values:. Defaults to whole
      axis: The `axis` parameter determines whether the axis of the plot should be displayed or not. If
    `axis` is set to "on", the axis will be displayed. If `axis` is set to "off", the axis will not be
    displayed. Defaults to on
      dpi: The `dpi` parameter determines the resolution of the plot. Defaults to 300

    Returns:
      a matplotlib Figure object.
    """
    _, ax = plt.subplots(figsize=(10, 10), dpi=dpi)
    crs = gdf.crs
    # Filter based on direction and level
    if "direction" in kwargs:
        gdf = gdf[gdf.direction_id == kwargs["direction"]].copy()
    if level == "whole":
        markersize = 20
        ax = gdf.plot(
            ax=ax,
            color="#34495e",
            linewidth=0.5,
            edgecolor="black",
            label="Bus network",
            zorder=1,
        )
    elif level == "route":
        markersize = 40
        assert "route" in kwargs, "Please provide a route_id in route attibute"
        gdf = gdf[gdf.route_id.isin(kwargs["route"])].copy()
    elif level == "segment":
        markersize = 60
        assert "segment" in kwargs, "Please provide a segment_id in segment attibute"
        gdf = gdf[gdf.segment_id.isin(kwargs["segment"])].copy()
    else:
        raise ValueError("level must be either whole, route, or segment")

    # Plot the spacings
    if "route" in kwargs:
        gdf = gdf[gdf.route_id.isin(kwargs["route"])].copy()
        if len(kwargs["route"]) > 1:
            ax = gdf.plot(
                ax=ax,
                linewidth=2,
                column="route_id",
                label="Route ID:" + str(kwargs["route"]),
                zorder=2,
                cmap="tab20",
                legend=True,
            )
        else:
            ax = gdf.plot(
                ax=ax,
                linewidth=2,
                color="#2ecc71",
                label="Route ID:" + str(kwargs["route"]),
                zorder=2,
            )
    if "segment" in kwargs:
        try:
            gdf = gdf[gdf.segment_id.isin(kwargs["segment"])].copy()
        except ValueError as e:
            raise ValueError(f"No such segment exists. Check if direction_id is incorrect {e}")
        ax = gdf.plot(
            ax=ax,
            linewidth=2.5,
            color="#000000",
            label="Segment ID: " + str(kwargs["segment"]),
            zorder=3,
        )
    if show_stops:
        geo_series = gdf.geometry.apply(lambda line: Point(line.coords[0]))
        geo_series = pd.concat([geo_series, gpd.GeoSeries(Point(gdf.iloc[-1].geometry.coords[-1]))])
        geo_series.set_crs(crs=gdf.crs).plot(
            ax=ax,
            color="#FFD700",
            edgecolor="#000000",
            linewidth=1,
            markersize=markersize,
            alpha=0.95,
            zorder=10,
        )

    if basemap:
        df = gpd.GeoDataFrame(gdf, crs=crs)
        cx.add_basemap(ax, crs=df.crs, source=map_provider, attribution_size=5)
    plt.axis(axis)
    if level != "segment":
        plt.legend(loc="best")
    else:
        ax.legend().set_visible(False)
    return ax


def view_spacings_interactive(
    gdf: gpd.GeoDataFrame,
    basemap: bool = True,
    show_stops: bool = False,
    level: str = "whole",
    **kwargs: Any,
) -> folium.Map:
    """
    Generates an interactive Folium map to visualize stop spacings.

    Parameters:
        gdf (gpd.GeoDataFrame): The GeoDataFrame containing the stop spacing data.
        basemap (bool, optional): Whether to add a basemap to the map. Defaults to True.
        show_stops (bool, optional): Whether to show the stops on the map. Defaults to False.
        level (str, optional): The level at which to filter the data. Can be 'whole', 'route', or 'segment'.
            Defaults to 'whole'.
        **kwargs: Additional keyword arguments for filtering the data based on level.

    Returns:
        folium.Map: The generated Folium map.

    Raises:
        AssertionError: If the required attributes for filtering the data are not provided.

    Example usage:
        gdf = gpd.GeoDataFrame(...)
        map = view_spacings_interactive(gdf, basemap=True, show_stops=True, level='route', route='123')
    """
    if "direction" in kwargs:
        gdf = gdf[gdf.direction_id == kwargs["direction"]].copy()
    # Convert CRS to EPSG:4326 if needed
    if gdf.crs.to_string() != "EPSG:4326":
        gdf = gdf.to_crs(epsg=4326)

    # Initialize Folium Map
    bounds = gdf.total_bounds
    map_center = [(bounds[1] + bounds[3]) / 2, (bounds[0] + bounds[2]) / 2]
    fmap = folium.Map(location=map_center, control_scale=True, zoom_start=12)

    # Filter and plot based on level
    if level == "route":
        assert "route" in kwargs, "Please provide a route_id in route attribute"
        gdf = gdf[gdf.route_id == kwargs["route"]].copy()
    elif level == "segment":
        assert "segment" in kwargs, "Please provide a segment_id in segment attribute"
        gdf = gdf[gdf.segment_id == kwargs["segment"]].copy()

    # Add lines to map
    tooltip = folium.GeoJsonTooltip(fields=["segment_id", "distance"])
    popup = folium.GeoJsonPopup(fields=gdf.drop(columns=["geometry"]).columns.tolist())

    def style_function(x: Any) -> dict[str, Any]:
        if "route" in kwargs:
            return {
                "color": (
                    "#2ecc71" if x["properties"]["route_id"] == kwargs["route"] else "#34495e"
                ),
                "weight": (5 if x["properties"]["route_id"] == kwargs["route"] else 2),
            }
        if "segment" in kwargs:
            return {
                "color": (
                    "#000000" if x["properties"]["segment_id"] == kwargs["segment"] else "#34495e"
                ),
                "weight": (5 if x["properties"]["segment_id"] == kwargs["segment"] else 2),
                "z_index": 1000,
            }
        return {"color": "#34495e", "weight": 2}

    folium.GeoJson(
        gdf, tooltip=tooltip, popup=popup, zoom_on_click=True, style_function=style_function
    ).add_to(fmap)

    # Show stops
    if show_stops:
        if "route" in kwargs:
            gdf = gdf[gdf.route_id == kwargs["route"]].copy()
        if "segment" in kwargs:
            gdf = gdf[gdf.segment_id == kwargs["segment"]].copy()
        stop_ids = {}
        for _, row in gdf.iterrows():
            stop_ids[row["stop_id1"]] = Point(row["geometry"].coords[0])
            stop_ids[row["stop_id2"]] = Point(row["geometry"].coords[-1])
        for stop_id, point in stop_ids.items():
            folium.CircleMarker(
                location=[point.y, point.x],
                radius=(6 if "segment" in kwargs else 4 if "route" in kwargs else 2),
                scale_radius=True,
                weight=1,
                fill_opacity=0.9,
                color="#000000",
                fill_color="#FFD700",
                fill=True,
                tooltip=str(stop_id),
            ).add_to(fmap)

    # Add basemap
    if basemap:
        folium.TileLayer("CartoDB positron", name="Light Map", control=False).add_to(fmap)

    return fmap


def increase_resolution(geom: LineString, spat_res: int = 5) -> LineString:
    """
    This function increases the resolution of a LineString geometry by adding
    points along the line at a specified spatial resolution.

    Args:
      geom: The input geometry that needs to be modified (in this case, a LineString).
      spat_res: spatial resolution, which is the desired distance between consecutive points
        on the LineString. If the distance between two consecutive points is greater than the
        spatial resolution, the function will add additional points to the LineString to
        increase its resolution. Defaults to 5

    Returns:
      a LineString object with increased resolution based on the input spatial resolution.
    """
    coords = geom.coords
    coord_pairs = np.array([coords[i : i + 2] for i in range(len(coords) - 1)])
    coord_dists = np.array(
        [geod.geometry_length(LineString(coords[i : i + 2])) for i in range(len(coords) - 1)]
    )
    new_ls = []
    for i, dists in enumerate(coord_dists):
        pair = coord_pairs[i]
        if dists > spat_res:
            factor = int(np.ceil(dists / spat_res))
            ret_points = [tuple(pair[0])]
            for j in range(1, factor):
                new_point = (
                    pair[0][0] + (pair[1][0] - pair[0][0]) * j / factor,
                    pair[0][1] + (pair[1][1] - pair[0][1]) * j / factor,
                )
                ret_points.append(new_point)
            for pt in ret_points:
                new_ls.append(pt)
        else:
            new_ls.append(tuple(pair[0]))
    new_ls.append(tuple(coord_pairs[-1][1]))
    return LineString(new_ls)


def ret_high_res_shape(
    shapes: gpd.GeoDataFrame, trips: pd.DataFrame, spat_res: int = 5
) -> gpd.GeoDataFrame:
    """
    This function increases the resolution of the geometries in a given dataframe of shapes by a
    specified spatial resolution.

    Args:
      shapes: a pandas DataFrame containing a column named 'geometry' that contains shapely geometry
    objects
      spat_res: spatial resolution, which is the size of each pixel or cell in a raster dataset. In this
    function, it is used to increase the resolution of the input shapes by creating more vertices in
    their geometries. The default value is 5, which means that the resolution will be increased by
    adding vertices. Defaults to 5

    Returns:
      a GeoDataFrame with the geometry column updated to have higher resolution shapes.
    """
    shape_ids = trips.shape_id.unique()
    shapes = shapes[shapes.shape_id.isin(shape_ids)].copy()
    high_res_shapes = [
        increase_resolution(row["geometry"], spat_res) for i, row in shapes.iterrows()
    ]
    shapes.geometry = high_res_shapes
    return shapes


def ret_high_res_shape_parallel(shapes: gpd.GeoDataFrame, spat_res: int = 5) -> gpd.GeoDataFrame:
    """
    This function increases the resolution of the geometries in a given dataframe of shapes by a
    specified spatial resolution.

    Args:
      shapes: a pandas DataFrame containing a column named 'geometry' that contains shapely geometry
    objects
      spat_res: spatial resolution, which is the size of each pixel or cell in a raster dataset. In this
    function, it is used to increase the resolution of the input shapes by creating more vertices in
    their geometries. The default value is 5, which means that the resolution will be increased by
    adding vertices. Defaults to 5

    Returns:
      a GeoDataFrame with the geometry column updated to have higher resolution shapes.
    """

    def process_shape(row: pd.core.series.Series) -> LineString:
        return increase_resolution(row["geometry"], spat_res)

    high_res_shapes = []
    with ThreadPoolExecutor(max_workers=None) as executor:
        high_res_shapes = list(executor.map(process_shape, shapes.to_dict("records")))

    shapes.geometry = high_res_shapes
    return shapes


def nearest_points(stop_df: gpd.GeoDataFrame, k_neighbors: int = 3) -> pd.DataFrame:
    """
    The function takes a dataframe of stops and snaps them to the nearest points on a line geometry,
    with an option to specify the number of nearest neighbors to consider.

    Args:
      stop_df: a pandas DataFrame containing information about stops along a set of trips, including the
    trip ID, the stop location (as a Shapely Point object), and the geometry of the trip (as a Shapely
    LineString object)
      k_neighbors: The number of nearest neighbors to consider when snapping stops to a line geometry.
    Default value is 3. Defaults to 3

    Returns:
      the stop_df dataframe with an additional column 'snap_start_id' which contains the indices of the
    nearest points on the trip route for each stop. If any trips failed to snap, they are excluded from
    the returned dataframe.
    """
    stop_df["snap_start_id"] = -1
    geo_const = 6371000 * np.pi / 180
    failed_trips = []
    count = 0
    total_trip_count = 0
    defective_trip_count = 0
    for name, group in stop_df.groupby("trip_id"):
        # print(name)
        count += 1
        total_trip_count += len(group)
        neighbors = k_neighbors
        geom_line = group["geometry"].iloc[0]
        # print(len(geom_line.coords))
        tree = cKDTree(data=np.array(geom_line.coords))
        stops = [x.coords[0] for x in group["start"]]
        if len(stops) <= 1:
            failed_trips.append(name)
            print("Excluding Trip: " + name + " because of too few stops")
            defective_trip_count += len(group)
            continue
        failed_trip = False
        solution_found = False
        while not solution_found:
            np_dist, np_inds = tree.query(stops, workers=-1, k=neighbors)
            # Approx distance in meters
            np_dist = np_dist * geo_const
            prev_point = min(np_inds[0])
            points = [prev_point]
            for i, nps in enumerate(np_inds[1:]):
                condition = (nps > prev_point) & (nps < max(np_inds[i + 1]))
                points_valid = nps[condition]
                if len(points_valid) > 0:
                    points_score = (np.power(points_valid - prev_point, 3)) * np.power(
                        np_dist[i + 1, condition], 1
                    )
                    prev_point = nps[condition][np.argmin(points_score)]
                    points.append(prev_point)
                else:
                    # No valid points found
                    if neighbors < len(stops):
                        neighbors = min(neighbors + 2, len(stops))
                        break
                    else:
                        failed_trips.append(name)
                        failed_trip = True
                        # Make this to exit the while loop
                        solution_found = True
                        print("Excluding Trip: " + name + " because of failed snap!")
                        defective_trip_count += len(group)
                        break
            if len(points) == len(stops):
                solution_found = True
        if len(points) != len(set(points)):
            print("Processing", count, len(stop_df.trip_id.unique()))
            print("Points defective")

        if not failed_trip:
            stop_df.loc[stop_df.trip_id == name, "snap_start_id"] = points

    print("Total trips processed: ", total_trip_count)
    if defective_trip_count > 0:
        print("Total defective trips: ", defective_trip_count)
        print(
            "Percentage defective trips: ",
            defective_trip_count / total_trip_count * 100,
        )
    stop_df = stop_df[~stop_df.trip_id.isin(failed_trips)].reset_index(drop=True)
    return stop_df


# def process_trip_group(
#     name: str, group: pd.core.groupby.DataFrameGroupBy, k_neighbors: int, geo_const: float
# ) -> Tuple:
#     neighbors = k_neighbors
#     geom_line = group["geometry"].iloc[0]
#     tree = cKDTree(data=np.array(geom_line.coords))
#     stops = [x.coords[0] for x in group["start"]]
#     n_stops = len(stops)
#     if n_stops <= 1:
#         return name, None, True  # Failed trip due to too few stops

#     failed_trip = False
#     solution_found = False
#     points = []
#     while not solution_found:
#         np_dist, np_inds = tree.query(stops, workers=-1, k=neighbors)
#         np_dist = np_dist * geo_const  # Approx distance in meters
#         prev_point = min(np_inds[0])
#         points = [prev_point]
#         for i, nps in enumerate(np_inds[1:]):
#             condition = (nps > prev_point) & (nps < max(np_inds[i + 1]))
#             points_valid = nps[condition]
#             if len(points_valid) > 0:
#                 points_score = np.power(points_valid - prev_point, 3) * np.power(
#                     np_dist[i + 1, condition], 1
#                 )
#                 prev_point = nps[condition][np.argmin(points_score)]
#                 points.append(prev_point)
#             else:
#                 # Capping the number of nearest neighbors to 11
#                 if neighbors < min(n_stops, 11):
#                     neighbors = min(neighbors + 2, n_stops)
#                     break
#                 else:
#                     failed_trip = True
#                     solution_found = True
#                     break
#         if len(points) == n_stops:
#             solution_found = True

#     if failed_trip:
#         return name, None, True
#     else:
#         return name, points, False


def process_trip_group(
    name: str, group: pd.core.groupby.DataFrameGroupBy, k_neighbors: int, geo_const: float
) -> Tuple:
    neighbors = k_neighbors
    geom_line = group["geometry"].iloc[0]
    tree = cKDTree(data=np.array(geom_line.coords))
    stops = [x.coords[0] for x in group["start"]]
    n_stops = len(stops)
    MAX_NEIGHBORS = min(n_stops, 9)
    if n_stops <= 1:
        return name, None, True  # Failed trip due to too few stops

    failed_trip = False
    solution_found = False
    points = []
    np_dist_all, np_inds_all = tree.query(stops, workers=-1, k=MAX_NEIGHBORS)
    np_dist_all = np_dist_all * geo_const  # Approx distance in meters
    while not solution_found:
        np_inds = np_inds_all[:, :neighbors]
        np_dist = np_dist_all[:, :neighbors]
        prev_point = min(np_inds[0])
        points = [prev_point]
        for i, nps in enumerate(np_inds[1:]):
            condition = (nps > prev_point) & (nps < max(np_inds[i + 1]))
            points_valid = nps[condition]
            if len(points_valid) > 0:
                points_score = np.power(points_valid - prev_point, 3) * np.power(
                    np_dist[i + 1, condition], 1
                )
                prev_point = nps[condition][np.argmin(points_score)]
                points.append(prev_point)
            else:
                # Capping the number of nearest neighbors to 11
                if neighbors < MAX_NEIGHBORS:
                    neighbors = min(neighbors + 2, n_stops)
                    break
                else:
                    failed_trip = True
                    solution_found = True
                    break
        if len(points) == n_stops:
            solution_found = True

    if failed_trip:
        return name, None, True
    else:
        return name, points, False


def nearest_points_parallel(stop_df: gpd.GeoDataFrame, k_neighbors: int = 5) -> pd.DataFrame:
    stop_df["snap_start_id"] = -1
    geo_const = 6371000 * np.pi / 180
    failed_trips = []
    defective_trip_count = 0
    with ThreadPoolExecutor(max_workers=None) as executor:
        results = executor.map(
            lambda x: process_trip_group(x[0], x[1], k_neighbors, geo_const),
            stop_df.groupby("trip_id"),
        )

    for name, points, failed in results:
        if failed:
            failed_trips.append(name)
        else:
            stop_df.loc[stop_df.trip_id == name, "snap_start_id"] = points
    defective_trip_count = (
        stop_df[stop_df.trip_id.isin(failed_trips)].groupby("trip_id").first().traversals.sum()
    )
    total_trip_count = len(stop_df)
    stop_df = stop_df[~stop_df.trip_id.isin(failed_trips)].reset_index(drop=True)

    print("Total trips processed:", total_trip_count)
    if defective_trip_count > 0:
        print("Total defective trips:", defective_trip_count)
        print(
            "Percentage defective trips:{:.2f}".format(
                defective_trip_count / total_trip_count * 100
            )
        )
    return stop_df


def view_heatmap(
    gdf: gpd.GeoDataFrame,
    column: str = "distance",
    cmap: Optional[str] = "RdYlBu",
    light_mode: bool = True,
    interactive: bool = False,
) -> Any:
    """
    Generates a heatmap visualization of a GeoDataFrame.

    Parameters:
        gdf (gpd.GeoDataFrame): The GeoDataFrame containing the data to be visualized.
        cmap (Optional[str], optional): The colormap to be used for the heatmap. Defaults to "RdYlBu".
        light_mode (bool, optional): Specifies whether to use a light mode basemap. Defaults to True.
        interactive (bool, optional): Specifies whether to generate an interactive map. Defaults to False.

    Returns:
        Any: The generated heatmap visualization.

    """
    df_filtered = gdf.copy()
    df_filtered[column] = pd.to_numeric(df_filtered[column])
    if column == "distance":
        MAX_RANGE = gdf["distance"].max()
        df_filtered = gdf[(gdf["distance"] >= 30)].copy()
        bins = [125, 200, 400, 600, 800, 1000, 1200, 1500, 2000, MAX_RANGE]
    else:
        df_filtered = df_filtered[(df_filtered[column] >= df_filtered[column].quantile(0.01))]
        df_filtered = df_filtered[(df_filtered[column] <= df_filtered[column].quantile(1 - 0.01))]
    if interactive:
        if column == "distance":
            fmap = df_filtered.explore(
                column=column,
                scheme="UserDefined",
                tooltip=["segment_id", "distance"],
                tiles="CartoDB Positron" if light_mode else "CartoDB Dark Matter",
                legend=True,
                cmap=cmap,  # YlOrRd
                classification_kwds=dict(bins=bins),
                legend_kwds=dict(colorbar=False, fmt="{:.0f}"),
                style_kwds=dict(opacity=0.75, fillOpacity=0.75),
                popup=True,
            )
        else:
            fmap = df_filtered.explore(
                column=column,
                cmap=cmap,  # YlOrRd
                tooltip=["segment_id", column],
                tiles="CartoDB Positron" if light_mode else "CartoDB Dark Matter",
                legend=True,
                style_kwds=dict(opacity=0.75, fillOpacity=0.75),
                popup=True,
                scheme="Quantiles",
                legend_kwds=dict(colorbar=False, fmt="{:.0f}"),
            )
        return fmap
    else:
        fig, ax = plt.subplots(figsize=(10, 8), dpi=300)
        if column == "distance":
            df_filtered.plot(
                column=column,
                scheme="UserDefined",
                cmap=cmap,  # YlOrRd
                kind="geo",
                ax=ax,
                legend=True,
                classification_kwds=dict(bins=bins),
                legend_kwds=dict(
                    fmt="{:.0f}", loc="upper left", bbox_to_anchor=(0, 1), interval=True
                ),
                alpha=0.75,
            )
        else:
            df_filtered.plot(
                column=column,
                cmap=cmap,  # YlOrRd
                kind="geo",
                ax=ax,
                legend=True,
                alpha=0.275,
                scheme="Quantiles",
            )
        map_provider = (
            cx.providers.CartoDB.Positron if light_mode else cx.providers.CartoDB.DarkMatter
        )
        cx.add_basemap(ax, crs=gdf.crs, source=map_provider, attribution_size=5)
        plt.axis("off")
        plt.close()
        return fig



import os
from typing import List, Optional, Set

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import LineString

from .geom_utils import (
    get_zone_epsg,
    make_gdf,
    nearest_points,
    nearest_points_parallel,
    ret_high_res_shape,
)
from .mobility import summary_stats_mobility
from .partridge_func import get_bus_feed
from .partridge_mod.gtfs import Feed
from .utils import download_write_file, export_segments, failed_pipeline, plot_hist


def merge_trip_geom(trip_df: pd.DataFrame, shape_df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    It takes a dataframe of trips and a dataframe of shapes, and returns a geodataframe of trips with
    the geometry of the shapes

    Args:
      trip_df: a dataframe of trips
      shape_df: a GeoDataFrame of the shapes.txt file

    Returns:
      A GeoDataFrame
    """
    # `direction_id` and `shape_id` are optional
    if "direction_id" in trip_df.columns:
        # Check is direction_ids are listed as null
        if trip_df["direction_id"].isnull().sum() == 0:
            grp = trip_df.groupby(["route_id", "shape_id", "direction_id"])
        else:
            grp = trip_df.groupby(["route_id", "shape_id"])
    else:
        grp = trip_df.groupby(["route_id", "shape_id"])
    trip_df = grp.first().reset_index()
    trip_df["traversals"] = grp.count().reset_index(drop=True)["trip_id"]
    subset_list = np.array(
        ["route_id", "trip_id", "shape_id", "service_id", "direction_id", "traversals"]
    )
    col_subset = subset_list[np.in1d(subset_list, trip_df.columns)]
    trip_df = trip_df[col_subset]
    trip_df = trip_df.dropna(how="all", axis=1)
    trip_df = shape_df.merge(trip_df, on="shape_id", how="left")
    return make_gdf(trip_df)


def make_segments_unique(df: gpd.GeoDataFrame, traversal_threshold: int = 1) -> gpd.GeoDataFrame:
    # Compute the number of unique rounded distances for each route_id and segment_id
    unique_counts = df.groupby(["route_id", "segment_id"])["distance"].apply(
        lambda x: x.round().nunique()
    )

    # Filter rows where unique count is greater than 1
    filtered_df = df[
        df.set_index(["route_id", "segment_id"]).index.isin(unique_counts[unique_counts > 1].index)
    ].copy()

    # Create a segment modification function
    def modify_segment(segment_id: str, count: int) -> str:
        seg_split = str(segment_id).split("-")
        return seg_split[0] + "-" + seg_split[1] + "-" + str(count + 1)

    # Apply the modification function to the segment_id
    filtered_df["modification"] = filtered_df.groupby(["route_id", "segment_id"]).cumcount()
    filtered_df["segment_id"] = filtered_df.apply(
        lambda row: modify_segment(row["segment_id"], row["modification"])
        if row["modification"] != 0
        else row["segment_id"],
        axis=1,
    )

    # Merge the modified segments back into the original DataFrame
    df = pd.concat([df[~df.index.isin(filtered_df.index)], filtered_df], ignore_index=True)

    # Aggregate traversals and filter by traversal threshold
    grp_again = df.groupby(["route_id", "segment_id"])
    df = grp_again.first().reset_index()
    df["traversals"] = grp_again["traversals"].sum().values
    df = df[df.traversals > traversal_threshold].reset_index(drop=True)
    return make_gdf(df)


def filter_stop_df(stop_df: pd.DataFrame, trip_ids: Set, stop_loc_df: pd.DataFrame) -> pd.DataFrame:
    """
    It takes a dataframe of stops and a list of trip IDs and returns a dataframe of stops that are in
    the list of trip IDs

    Args:
      stop_df: the dataframe of all stops
      trip_ids: a list of trip_ids that you want to filter the stop_df by

    Returns:
      A dataframe with the trip_id, s top_id, and stop_sequence for the trips in the trip_ids list.
    """
    missing_stop_locs = set(stop_df.stop_id) - set(stop_loc_df.stop_id)
    if len(missing_stop_locs) > 0:
        print("Missing stop locations for:", missing_stop_locs)
        missing_trips = stop_df[stop_df.stop_id.isin(missing_stop_locs)].trip_id.unique()
        for trip in missing_trips:
            trip_ids.discard(trip)
            print(
                "Removed the trip_id:", trip, "as stop locations are missing for stops in the trip"
            )
    # Filter the stop_df to only include the trip_ids in the trip_ids list
    stop_df = stop_df[stop_df.trip_id.isin(trip_ids)].reset_index(drop=True)
    stop_df = stop_df.sort_values(["trip_id", "stop_sequence"]).reset_index(drop=True)
    stop_df["main_index"] = stop_df.index
    stop_df_grp = stop_df.groupby("trip_id")
    drop_inds = []
    # To eliminate deadheads
    if "pickup_type" in stop_df.columns:
        grp_f = stop_df_grp.first()
        drop_inds.append(grp_f.loc[grp_f["pickup_type"] == 1, "main_index"])
    if "drop_off_type" in stop_df.columns:
        grp_l = stop_df_grp.last()
        drop_inds.append(
            grp_l.loc[grp_l["drop_off_type"] == 1, "main_index"]
        )  # Fixed the variable name from grp_f to grp_l
    if len(drop_inds) > 0 and len(drop_inds[0]) > 0:
        stop_df = stop_df[~stop_df["main_index"].isin(drop_inds)].reset_index(drop=True)
    stop_df = stop_df[["trip_id", "stop_id", "stop_sequence", "arrival_time"]]

    stop_df = stop_df.sort_values(["trip_id", "stop_sequence"]).reset_index(drop=True)
    return stop_df


def merge_stop_geom(stop_df: pd.DataFrame, stop_loc_df: pd.DataFrame) -> gpd.GeoDataFrame:
    """
    > Merge the stop_loc_df with the stop_df, and then convert the result to a GeoDataFrame

    Args:
      stop_df: a dataframe of stops
      stop_loc_df: a GeoDataFrame of the stops

    Returns:
      A GeoDataFrame
    """
    stop_df["start"] = stop_df.copy().merge(stop_loc_df, how="left", on="stop_id")["geometry"]
    return stop_df


def create_segments(stop_df: gpd.GeoDataFrame, parallel: bool = False) -> pd.DataFrame:
    """
    This function creates segments between stops based on their proximity and returns a GeoDataFrame.

    Args:
      stop_df: A pandas DataFrame containing information about stops on a transit network, including
    their stop_id, coordinates, and trip_id.

    Returns:
      a GeoDataFrame with segments created from the input stop_df.
    """
    if parallel:
        stop_df = nearest_points_parallel(stop_df)
    else:
        stop_df = nearest_points(stop_df)
    stop_df = stop_df.rename({"stop_id": "stop_id1", "arrival_time": "arrival_time1"}, axis=1)
    grp = (
        pd.DataFrame(stop_df).groupby("trip_id", group_keys=False).shift(-1).reset_index(drop=True)
    )
    stop_df[["stop_id2", "end", "snap_end_id", "arrival_time2"]] = grp[
        ["stop_id1", "start", "snap_start_id", "arrival_time1"]
    ]
    stop_df["segment_id"] = stop_df.apply(
        lambda row: str(row["stop_id1"]) + "-" + str(row["stop_id2"]) + "-1", axis=1
    )
    stop_df = stop_df.dropna().reset_index(drop=True)
    stop_df.snap_end_id = stop_df.snap_end_id.astype(int)
    stop_df = stop_df[stop_df["snap_end_id"] > stop_df["snap_start_id"]].reset_index(drop=True)
    stop_df["geometry"] = stop_df.apply(
        lambda row: LineString(
            row["geometry"].coords[row["snap_start_id"] : row["snap_end_id"] + 1]
        ),
        axis=1,
    )
    return stop_df


def process_feed_stops(feed: Feed) -> gpd.GeoDataFrame:
    """
    It takes a GTFS feed, merges the trip and shape data, filters the stop_times data to only include
    the trips that are in the feed, merges the stop_times data with the stop data, creates a segment for
    each stop pair, gets the EPSG zone for the feed, creates a GeoDataFrame, and calculates the length
    of each segment

    Args:
      feed: a GTFS feed object
      max_spacing: the maximum distance between stops in meters. If a stop is more than this distance
    from the previous stop, it will be dropped.

    Returns:
      A GeoDataFrame with the following columns:
    """
    trip_df = merge_trip_geom(feed.trips, feed.shapes)
    trip_ids = set(trip_df.trip_id.unique())
    stop_loc_df = feed.stops[["stop_id", "geometry"]]
    stop_df = filter_stop_df(feed.stop_times, trip_ids, stop_loc_df)
    stop_df = merge_stop_geom(stop_df, stop_loc_df)
    stop_df = stop_df.merge(trip_df, on="trip_id", how="left")
    stops = stop_df.groupby("shape_id").count().reset_index()["geometry"]
    stop_df = stop_df.groupby("shape_id").first().reset_index()
    stop_df["n_stops"] = stops
    epsg_zone = get_zone_epsg(stop_df)
    if epsg_zone is not None:
        stop_df["distance"] = stop_df.geometry.to_crs(epsg_zone).length
        stop_df["mean_distance"] = stop_df["distance"] / stop_df["n_stops"]
    return make_gdf(stop_df)


def process_feed(
    feed: Feed, parallel: bool = False, max_spacing: Optional[float] = None
) -> gpd.GeoDataFrame:
    """
    The function `process_feed` takes a feed and optional maximum spacing as input, performs various
    data processing and filtering operations on the feed, and returns a GeoDataFrame containing the
    processed data.

    Args:
      feed: The `feed` parameter is a data structure that contains information about a transit network.
    It likely includes data such as shapes (geometric representations of routes), trips (sequences of
    stops), stop times (arrival and departure times at stops), and stops (locations of stops).
      [Optional] max_spacing: The `max_spacing` parameter is an optional parameter that specifies the maximum
    distance between stops. If provided, the function will filter out stops that are farther apart than
    the specified maximum spacing.

    Returns:
      A GeoDataFrame containing information about the stops and segments in the feed with segments smaller than the max_spacing values.
    """
    # Set a Spatial Resolution and increase the resolution of the shapes
    # shapes = ret_high_res_shape_parallel(feed.shapes, spat_res=5)
    ## Note: Currently, the parallel version of the function is not working as expected and is slower than the non-parallel version
    shapes = ret_high_res_shape(feed.shapes, feed.trips, spat_res=5)
    trip_df = merge_trip_geom(feed.trips, shapes)
    trip_ids = set(trip_df.trip_id.unique())
    stop_loc_df = feed.stops[["stop_id", "geometry"]]
    stop_df = filter_stop_df(feed.stop_times, trip_ids, stop_loc_df)
    stop_df = merge_stop_geom(stop_df, stop_loc_df)
    stop_df = stop_df.merge(trip_df, on="trip_id", how="left")
    stop_df = create_segments(stop_df, parallel=parallel)
    stop_df = make_gdf(stop_df)
    epsg_zone = get_zone_epsg(stop_df)
    if epsg_zone is not None:
        stop_df["distance"] = stop_df.set_geometry("geometry").to_crs(epsg_zone).geometry.length
        stop_df["distance"] = stop_df["distance"].round(2)  # round to 2 decimal places
    stop_df["traversal_time"] = stop_df["arrival_time2"] - stop_df["arrival_time1"]
    stop_df = stop_df[stop_df["traversal_time"] > 0].reset_index(drop=True)
    stop_df["speed"] = stop_df["distance"] / stop_df["traversal_time"]
    stop_df = make_segments_unique(stop_df, traversal_threshold=1)
    subset_list = np.array(
        [
            "segment_id",
            "route_id",
            "direction_id",
            "trip_id",
            "traversals",
            "distance",
            "stop_id1",
            "stop_id2",
            "traversal_time",
            "speed",
            "geometry",
        ]
    )
    col_subset = subset_list[np.in1d(subset_list, stop_df.columns)]
    stop_df = stop_df[col_subset]
    if max_spacing is not None:
        stop_df = stop_df[stop_df["distance"] <= max_spacing]
    return make_gdf(stop_df)


def inspect_feed(feed: Feed) -> str:
    """
    It checks to see if the feed has any bus routes and if it has a `shape_id` column in the `trips`
    table

    Args:
      feed: The feed object that you want to inspect.

    Returns:
      A message
    """
    message = "Valid GTFS Feed"
    if len(feed.stop_times) == 0:
        message = "No Bus Routes in "
    if "shape_id" not in feed.trips.columns:
        message = "Missing `shape_id` column in "
    return message


def get_gtfs_segments(
    path: str,
    agency_id: Optional[str] = None,
    threshold: Optional[int] = 1,
    max_spacing: Optional[float] = None,
    parallel: bool = False,
) -> gpd.GeoDataFrame:
    """
    The function `get_gtfs_segments` takes a path to a GTFS feed file, an optional agency name, a
    threshold value, and an optional maximum spacing value, and returns processed GTFS segments.

    Args:
      path: The path parameter is the file path to the GTFS (General Transit Feed Specification) data.
    This is the data format used by public transportation agencies to provide schedule and geographic
    information about their services.
      [Optional] agency_id: The agency_id of the transit agency for which you want to retrieve the bus feed. If this
    parameter is not provided, the function will retrieve the bus feed for all transit agencies. You can pass
    a list of agency_ids to retrieve the bus feed for multiple transit agencies.
      [Optional] threshold: The threshold parameter is used to filter out bus trips that have fewer stops than the
    specified threshold. Trips with fewer stops than the threshold will be excluded from the result.
    Defaults to 1
      [Optional] max_spacing: The `max_spacing` parameter is used to specify the maximum distance between two
    consecutive stops in a segment. If the distance between two stops exceeds the `max_spacing` value,
    the segment is split into multiple segments.

    Returns:
      A GeoDataFrame containing information about the stops and segments in the feed with segments
      smaller than the max_spacing values. Each row contains the following columns:
      - segment_id: the segment's identifier, produced by gtfs-segments
      - stop_id1: The `stop_id` identifier of the segment's beginning stop.
        The identifier is the same one the agency has chosen in the stops.txt file of its GTFS package.
      - stop_id2: The `stop_id` identifier of the segment's ending stop.
      - route_id: The same route ID listed in the agency's routes.txt file.
      - direction_id: The route's direction identifier.
      - traversals: The number of times the indicated route traverses the segment during the "measurement interval."
        The "measurement interval" chosen is the busiest day in the GTFS schedule: the day which has the most bus services running.
      - distance: The length of the segment in meters.
      - geometry: The segment's LINESTRING (a format for encoding geographic paths).
        All geometries are re-projected onto Mercator (EPSG:4326/WGS84) to maintain consistency.
    """
    feed = get_bus_feed(path, agency_id=agency_id, threshold=threshold, parallel=parallel)
    df = process_feed(feed, parallel=parallel)
    if max_spacing is not None:
        print("Using max_spacing {:.0f} to filter segments".format(max_spacing))
        df = df[df["distance"] <= max_spacing]
    return df


def pipeline_gtfs(filename: str, url: str, bounds: List, max_spacing: float) -> str:
    """
    It takes a GTFS file, downloads it, reads it, processes it, and then outputs a bunch of files.

    Let's go through the function step by step.

    First, we define the function and give it a name. We also give it a few arguments:

    - filename: the name of the file we want to save the output to.
    - url: the url of the GTFS file we want to download.
    - bounds: the bounding box of the area we want to analyze.
    - max_spacing: the maximum spacing we want to analyze.

    We then create a folder to save the output to.

    Next, we download the GTFS file and save it to the folder we just created.

    Then, we read the GTFS file using the `get_bus_feed` function.

    Args:
      filename: the name of the file you want to save the output to
      url: the url of the GTFS file
      bounds: the bounding box of the area you want to analyze. This is in the format
    [min_lat,min_lon,max_lat,max_lon]
      max_spacing: The maximum distance between stops that you want to consider.

    Returns:
      Success or Failure of the pipeline
    """
    folder_path = os.path.join("output_files", filename)
    gtfs_file_loc = download_write_file(url, folder_path)

    # read file using GTFS Fucntions
    feed = get_bus_feed(gtfs_file_loc)
    # Remove Null entries
    message = inspect_feed(feed)
    if message != "Valid GTFS Feed":
        return failed_pipeline(message, filename, folder_path)

    df = process_feed(feed)
    df_sub = df[df["distance"] < 3000].copy().reset_index(drop=True)
    if len(df_sub) == 0:
        return failed_pipeline("Only Long Bus Routes in ", filename, folder_path)
    # Output files and Stats
    summary_stats_mobility(df, folder_path, filename, url, bounds, max_spacing, export=True)

    plot_hist(
        df,
        file_path=os.path.join(folder_path, "spacings.png"),
        title=filename.split(".")[0],
        max_spacing=max_spacing,
        save_fig=True,
    )
    export_segments(
        df, os.path.join(folder_path, "geojson"), output_format="geojson", geometry=True
    )
    export_segments(
        df,
        os.path.join(folder_path, "spacings_with_geometry"),
        output_format="csv",
        geometry=True,
    )
    export_segments(df, os.path.join(folder_path, "spacings"), output_format="csv", geometry=False)
    return "Success for " + filename



import datetime
from functools import lru_cache
from typing import Any, Dict, List

import numpy as np
import pandas as pd

try:
    import geopandas as gpd
    from shapely.geometry import LineString
except ImportError as impexc:
    print(impexc)
    print("You must install GeoPandas to use this module.")
    raise

DATE_FORMAT = "%Y%m%d"


# Why 2^18? See https://git.io/vxB2P.
@lru_cache(maxsize=2**18)
def parse_time(val: str) -> int:
    """
    The function `parse_time` takes a string representing a time value in the format "hh:mm:ss" and
    returns the equivalent time in seconds as a numpy int, or returns the input value if it is
    already a numpy int or int.

    Args:
      val (str): The parameter `val` is a string representing a time value in the format "hh:mm:ss".

    Returns:
      a value of type np.float64.
    """
    if val is np.nan or isinstance(val, int):
        return val
    val = val.strip()
    h, m, s = val.split(":")
    ssm = int(h) * 3600 + int(m) * 60 + int(s)
    if val == "":
        return np.nan

    # pandas doesn't have a NaN int, use floats
    # return np.float64(ssm)
    return ssm


@lru_cache(maxsize=2**18)
def parse_date(val: str) -> datetime.date:
    """
    The function `parse_date` takes a string or a `datetime.date` object as input and returns a
    `datetime.date` object.

    Args:
      val (str): The `val` parameter is a string representing a date.

    Returns:
      a `datetime.date` object.
    """
    if isinstance(val, datetime.date):
        return val
    return datetime.datetime.strptime(val, DATE_FORMAT).date()


@lru_cache(maxsize=2**18)
def parse_numeric(val: Any) -> float:
    try:
        return float(val)
    except ValueError:
        return np.nan


vparse_numeric = np.vectorize(parse_numeric)
vparse_time = np.vectorize(parse_time)
vparse_date = np.vectorize(parse_date)

DEFAULT_CRS = "EPSG:4326"


def build_shapes(df: pd.DataFrame) -> gpd.GeoDataFrame:
    """
    The function takes a pandas DataFrame containing shape points and returns a GeoDataFrame with
    shape IDs and corresponding geometries.

    Args:
      df (pd.DataFrame): The parameter `df` is a pandas DataFrame that contains information
    about shapes. It is expected to have the following columns:

    Returns:
      a GeoDataFrame object.
    """
    if df.empty:
        return gpd.GeoDataFrame({"shape_id": [], "geometry": []})

    data: Dict[str, List] = {"shape_id": [], "geometry": []}
    for shape_id, shape in df.sort_values("shape_pt_sequence").groupby("shape_id"):
        data["shape_id"].append(shape_id)
        data["geometry"].append(LineString(list(zip(shape.shape_pt_lon, shape.shape_pt_lat))))

    return gpd.GeoDataFrame(data, crs=DEFAULT_CRS)


def build_stops(df: pd.DataFrame) -> gpd.GeoDataFrame:
    """
    The function `build_stops` takes a pandas DataFrame `df` and returns a GeoDataFrame with the
    same data but with a new geometry column created from the `stop_lon` and `stop_lat` columns.

    Args:
      df (pd.DataFrame): Pandas DataFrame that contains information about stops. It is expected to
      have columns named "stop_lon" and "stop_lat" which represent the longitude and
      latitude coordinates of each stop, respectively. The DataFrame may also contain other columns
      with additional information about the stops.

    Returns:
    GeoDataFrame with the geometry column containing points created from the stop_lon and stop_lat
    columns of the input DataFrame. The stop_lon and stop_lat columns are then dropped from the
    DataFrame before returning the final GeoDataFrame.
    """
    if df.empty:
        return gpd.GeoDataFrame(df, geometry=[], crs=DEFAULT_CRS)

    df = gpd.GeoDataFrame(
        df, crs=DEFAULT_CRS, geometry=gpd.points_from_xy(df.stop_lon, df.stop_lat)
    )

    df.drop(["stop_lon", "stop_lat"], axis=1, inplace=True)

    return gpd.GeoDataFrame(df, crs=DEFAULT_CRS)


def transforms_dict() -> Dict[str, Dict[str, Any]]:
    """
    The function `transforms_dict` returns a dictionary that specifies the required columns and
    converters for each file in a transit data feed.

    Returns:
      a dictionary containing information about various text files and their required columns and
    converters.
    """
    return_dict = {
        "agency.txt": {
            "usecols": {
                "agency_name": "str",
                "agency_url": "str",
                "agency_timezone": "str",
                "agency_lang": "str",
                "agency_phone": "int",
                "agency_fare_url": "str",
                "agency_email": "str",
            },
            "required_columns": ("agency_name", "agency_url", "agency_timezone"),
        },
        "calendar.txt": {
            "usecols": {
                "service_id": "str",
                "start_date": "str",
                "end_date": "str",
                "monday": "bool",
                "tuesday": "bool",
                "wednesday": "bool",
                "thursday": "bool",
                "friday": "bool",
                "saturday": "bool",
                "sunday": "bool",
            },
            "converters": {
                "start_date": vparse_date,
                "end_date": vparse_date,
                "monday": vparse_numeric,
                "tuesday": vparse_numeric,
                "wednesday": vparse_numeric,
                "thursday": vparse_numeric,
                "friday": vparse_numeric,
                "saturday": vparse_numeric,
                "sunday": vparse_numeric,
            },
            "required_columns": (
                "service_id",
                "monday",
                "tuesday",
                "wednesday",
                "thursday",
                "friday",
                "saturday",
                "sunday",
                "start_date",
                "end_date",
            ),
        },
        "calendar_dates.txt": {
            "usecols": {"service_id": "str", "date": "str", "exception_type": "int8"},
            "converters": {
                "date": vparse_date,
                "exception_type": vparse_numeric,
            },
            "required_columns": ("service_id", "date", "exception_type"),
        },
        "fare_attributes.txt": {
            "usecols": {
                "fare_id": "str",
                "price": "float",
                "currency_type": "str",
                "payment_method": "str",
                "transfers": "str",
                "transfer_duration": "float16",
            },
            "converters": {
                "price": vparse_numeric,
                "payment_method": vparse_numeric,
                "transfer_duration": vparse_numeric,
            },
            "required_columns": (
                "fare_id",
                "price",
                "currency_type",
                "payment_method",
                "transfers",
            ),
        },
        "fare_rules.txt": {
            "usecols": {
                "fare_id": "str",
                "route_id": "str",
                "origin_id": "str",
                "destination_id": "str",
                "contains_id": "str",
            },
            "required_columns": ("fare_id",),
        },
        "feed_info.txt": {
            "usecols": {
                "feed_publisher_name": "str",
                "feed_publisher_url": "str",
                "feed_lang": "str",
                "feed_start_date": "str",
                "feed_end_date": "str",
            },
            "converters": {
                "feed_start_date": vparse_date,
                "feed_end_date": vparse_date,
            },
            "required_columns": (
                "feed_publisher_name",
                "feed_publisher_url",
                "feed_lang",
            ),
        },
        "frequencies.txt": {
            "usecols": {
                "trip_id": "str",
                "start_time": "float32",
                "end_time": "float32",
                "headway_secs": "float32",
                "exact_times": "bool",
            },
            "converters": {
                "headway_secs": vparse_numeric,
                "exact_times": vparse_numeric,
                "start_time": vparse_time,
                "end_time": vparse_time,
            },
            "required_columns": (
                "trip_id",
                "start_time",
                "end_time",
                "headway_secs",
            ),
        },
        "routes.txt": {
            "usecols": {
                "route_id": "str",
                "route_short_name": "str",
                "route_long_name": "str",
                "route_type": "int8",
                # "route_color": "str",
                # "route_text_color": "str",
            },
            "converters": {
                "route_type": vparse_numeric,
            },
            "required_columns": (
                "route_id",
                "route_short_name",
                "route_long_name",
                "route_type",
            ),
        },
        "shapes.txt": {
            "usecols": {
                "shape_id": "str",
                "shape_pt_lat": "float32",
                "shape_pt_lon": "float32",
                "shape_pt_sequence": "int16",
                # "shape_dist_traveled":"float32",
            },
            "converters": {
                "shape_pt_lat": vparse_numeric,
                "shape_pt_lon": vparse_numeric,
                "shape_pt_sequence": vparse_numeric,
                # "shape_dist_traveled": vparse_numeric,
            },
            "required_columns": (
                "shape_id",
                "shape_pt_lat",
                "shape_pt_lon",
                "shape_pt_sequence",
            ),
            "transformations": [build_shapes],
        },
        "stops.txt": {
            "usecols": {
                "stop_id": "str",
                "stop_name": "str",
                "stop_lat": "float32",
                "stop_lon": "float32",
                # "location_type": "int8",
                # "wheelchair_boarding":"int8",
                # "timepoint":"bool",
            },
            "converters": {
                "stop_lat": vparse_numeric,
                "stop_lon": vparse_numeric,
                # "location_type": vparse_numeric,
                "wheelchair_boarding": vparse_numeric,
                "timepoint": vparse_numeric,
            },
            "required_columns": (
                "stop_id",
                "stop_name",
                "stop_lat",
                "stop_lon",
            ),
            "transformations": [build_stops],
        },
        "stop_times.txt": {
            "usecols": {
                "trip_id": "str",
                "arrival_time": "float32",
                # "departure_time",
                "stop_id": "str",
                "stop_sequence": "int16",
                "pickup_type": "int8",
                "drop_off_type": "int8",
                # "shape_dist_traveled",
                # "timepoint",
            },
            "converters": {
                "arrival_time": vparse_time,
                "departure_time": vparse_time,
                "pickup_type": vparse_numeric,
                "drop_off_type": vparse_numeric,
                # "shape_dist_traveled": vparse_numeric,
                "stop_sequence": vparse_numeric,
                # "timepoint": vparse_numeric,
            },
            "required_columns": (
                "trip_id",
                "arrival_time",
                "departure_time",
                "stop_id",
                "stop_sequence",
            ),
        },
        "transfers.txt": {
            "usecols": ["from_stop_id", "to_stop_id", "transfer_type", "min_transfer_time"],
            "converters": {
                "transfer_type": vparse_numeric,
                "min_transfer_time": vparse_numeric,
            },
            "required_columns": ("from_stop_id", "to_stop_id", "transfer_type"),
        },
        "trips.txt": {
            "usecols": {
                "route_id": "str",
                "shape_id": "str",
                "service_id": "str",
                "trip_id": "str",
                "direction_id": "bool",
                # "wheelchair_accessible": "int8",
                # "bikes_allowed":"int8",
            },
            "converters": {
                "direction_id": vparse_numeric,
                # "wheelchair_accessible": vparse_numeric,
                "bikes_allowed": vparse_numeric,
            },
            "required_columns": ("route_id", "service_id", "trip_id"),
        },
    }
    return return_dict



from datetime import timedelta
from typing import Any, Dict, List

import numpy as np
import pandas as pd
import utm
from numpy.typing import NDArray
from shapely.geometry import LineString

from .geom_utils import code
from .partridge_mod.gtfs import Feed


def get_zone_epsg_from_ls(geom: LineString) -> int:
    """
    > The function takes a dataframe with a geometry column and returns the EPSG code for the UTM zone
    that contains the geometry

    Args:
      stop_df: a dataframe with a geometry column

    Returns:
      The EPSG code for the UTM zone that the stop is in.
    """
    lon = geom.coords[0][0]
    lat = geom.coords[0][1]
    zone = utm.from_latlon(lat, lon)
    return code(zone, lat)


def get_sec(time_str: str) -> int:
    """
    It takes a string in the format of hh:mm:ss and returns the number of seconds

    Args:
      time_str: The time string to convert to seconds.

    Returns:
      the total number of seconds in the time string.
    """
    h, m, s = time_str.split(":")
    return int(h) * 3600 + int(m) * 60 + int(s)


def get_trips_len(df: pd.DataFrame, time: int) -> int:
    """
    > It returns the number of trips that are currently active at a given time

    Args:
      df: the dataframe of trips
      time: the time in seconds

    Returns:
      The number of trips that are active at a given time.
    """
    return len(df[(df.start_time <= time) & (df.end_time >= time)].trip_id.unique())


def get_trips(df: pd.DataFrame, time: int) -> pd.DataFrame:
    """
    > Return all trips that are currently active at a given time

    Args:
      df: the dataframe containing the trips
      time: the time in seconds

    Returns:
      A dataframe of all trips that start before the time and end after the time.
    """
    return df[(df.start_time <= time) & (df.end_time >= time)]


def get_peak_time(df: pd.DataFrame) -> List:
    """
    The function `get_peak_time` takes a dataframe of bus trips and returns the number of buses and the
    time at which the most buses are running.

    Args:
      df: The parameter `df` is a dataframe that contains information about buses. It likely has columns
    such as `start_time` and `end_time` that represent the start and end times of each bus trip.

    Returns:
      a list containing the number of buses and the time at which the most buses are running.
    """
    best = 0
    peak_time = 0
    start_time = int(min(df.start_time))
    end_time = int(max(df.end_time))
    for time in range(start_time, end_time, 60):
        no_buses = get_trips_len(df, time)
        if no_buses >= best:
            peak_time = str(timedelta(seconds=time))
            best = no_buses
    return [best, peak_time]


def get_service_length(df: pd.DataFrame) -> int:
    """
    The function takes a GTFS feed and a route_id, sorts the dataframe by stop_sequence, adjusts the
    shape_dist_traveled values if necessary, and returns the maximum shape_dist_traveled value.

    Args:
      df: The parameter `df` is a dataframe containing stop times data from a GTFS feed.

    Returns:
      The maximum value of the 'shape_dist_traveled' column in for the trip.
    """
    df = df.sort_values(["stop_sequence"])
    if df.iloc[0]["pickup_type"] == 1:
        sp_dist = df.iloc[1]["shape_dist_traveled"]
        df.shape_dist_traveled = df.shape_dist_traveled - sp_dist
        df.drop(index=df.iloc[0, :].index.tolist(), inplace=True, axis=0)
    return df.shape_dist_traveled.max()


def get_route_grp(route_df: pd.DataFrame) -> pd.DataFrame:
    """
    It takes a dataframe of route information and returns a dataframe of route information with the
    first and last stop times for each trip

    Args:
      route_df: The dataframe containing the route information

    Returns:
      A dataframe with the first and last stop of each trip.
    """
    route_df = route_df.sort_values(["stop_sequence"])
    route_df_grp = route_df.groupby(["trip_id"]).first()
    route_df_grp["start_time"] = route_df.groupby(["trip_id"]).first().arrival_time
    route_df_grp["end_time"] = route_df.groupby(["trip_id"]).last().arrival_time
    route_df_grp = route_df_grp.reset_index()
    col_filter = np.array(
        [
            "trip_id",
            "route_id",
            "direction_id",
            "start_time",
            "end_time",
            "pickup_type",
            "drop_off_type",
            "shape_dist_traveled",
        ]
    )
    col_subset = col_filter[np.in1d(col_filter, route_df_grp.columns)]
    return route_df_grp[col_subset]


def get_all_peak_times(df_dir: pd.DataFrame) -> Dict[str, NDArray[Any]]:
    """
    It takes a dataframe of bus trips and returns the peak number of buses and the time of the peak

    Args:
      df: the dataframe of the route you want to get the peak times for

    Returns:
      A dictionary with the peak number of buses in each direction, and the peak times.
    """
    best = 0
    peak_times = []
    df = get_route_grp(df_dir)
    start_time = int(min(df.start_time))
    end_time = int(max(df.end_time))
    for time in range(start_time, end_time, 60):
        no_buses = get_trips_len(df, time)
        if no_buses == best:
            peak_times.append(time)
        if no_buses > best:
            peak_times = [time]
    peak_time_condensed = np.array([str(timedelta(seconds=peak_times[0]))], dtype="object")
    for i, time in enumerate(peak_times):
        if i > 0:
            if (time - peak_times[i - 1]) != 60:
                time = str(timedelta(seconds=time))
                peak_time_condensed[-1] = (
                    str(peak_time_condensed[-1]) + "-" + str(timedelta(seconds=peak_times[i - 1]))
                )
                peak_time_condensed = np.append(peak_time_condensed, str(time))
            else:
                if i == len(peak_times) - 1:
                    peak_time_condensed[-1] = str(
                        peak_time_condensed[-1] + "-" + str(timedelta(seconds=peak_times[i]))
                    )
    return {"peak_buses": peak_time_condensed}


def get_average_headway(df_dir: pd.DataFrame) -> Dict[str, float]:
    """
    For each route, find the shape with the most trips, then find the stop with the most
    trips on that shape, then find the average headway for that stop

    Args:
      df_route: a dataframe containing the route you want to analyze

    Returns:
      A dictionary with the average headway for each direction.
    """
    hdw_0 = np.array([0])
    if len(df_dir) > 1:
        shape_0 = df_dir.groupby("shape_id").count().trip_id.idxmax()
        df_dir = df_dir[df_dir.shape_id == shape_0]
        stop_id0 = df_dir.stop_id.unique()[0]
        hdw_0 = (
            df_dir[df_dir.stop_id == stop_id0]
            .sort_values(["arrival_time"])
            .arrival_time.astype(int)
            .diff()
        )
    return {"headway": np.round(hdw_0[hdw_0 <= 3 * 60 * 60].mean() / 3600, 2)}


def get_average_speed(df_dir: pd.DataFrame, route_dict: dict) -> Dict[str, float]:
    """
    It takes a dataframe of a route and a dictionary of route information and returns a dictionary of
    average speeds for each direction

    Args:
      df_route: a dataframe of the route
      route_dict: a dictionary containing the route length and total time for each direction

    Returns:
      A dictionary with the average speed for each direction.
    """
    ret_dict = {}
    if len(df_dir) > 1:
        ret_dict["average_speed"] = np.round(
            route_dict["route_length"] / route_dict["total_time"], 2
        )
    return ret_dict


def get_route_time(df_dir: pd.DataFrame) -> Dict[str, float]:
    """
    For each route, find the shape_id that has the most trips, then find the trip_id that has that
    shape_id, then find the arrival_time of the first and last stop of that trip_id, then subtract the
    two to get the total time of the route

    Args:
      df_route: a dataframe of a route

    Returns:
      A dictionary with the total time for each direction.
    """
    time_0 = 0
    if len(df_dir) > 1:
        shape_0 = df_dir.groupby("shape_id").count().trip_id.idxmax()
        trip_0 = df_dir[df_dir.trip_id == df_dir[df_dir.shape_id == shape_0].trip_id.unique()[0]]
        time_0 = trip_0.arrival_time.max() - trip_0.arrival_time.min()
    return {"total_time": np.round(time_0 / 3600, 2) if time_0 != 0 else 0}


def get_bus_spacing(route_dict: dict) -> Dict[str, float]:
    """
    It takes a dataframe of a route and a dictionary of route information and returns a dictionary of
    the minimum spacing between buses on the route

    Args:
      df_route: a dataframe of all the trips for a given route
      route_dict: a dictionary containing the route length and number of buses for each direction

    Returns:
      A dictionary with the keys 'spacing dir 0' and 'spacing dir 1'
    """
    return {"bus_spacing": np.round(route_dict["route_length"] / route_dict["n_bus_avg"], 3)}


def average_active_buses(df_dir: pd.DataFrame) -> Dict[str, np.floating[Any]]:
    """
    Calculate the average number of active buses per time interval.

    Args:
      df_dir (pd.DataFrame): The input DataFrame containing bus data.

    Returns:
      Dict[str, np.floating[Any]]: A dictionary with the average number of active buses.

    """
    n_buses = []
    df = get_route_grp(df_dir)
    start_time = int(min(df.start_time))
    end_time = int(max(df.end_time))
    for time in range(start_time, end_time, 5 * 60):
        no_buses = get_trips_len(df, time)
        n_buses.append(no_buses)
    n_buses = np.array(n_buses)
    return {"n_bus_avg": np.round(np.mean(n_buses[n_buses > 0]), 3)}


def get_stop_spacing(df_dir: pd.DataFrame, route_dict: dict) -> Dict[str, float]:
    """
    For each route, find the shape with the most trips, and then find the number of stops on that trip.
    Then divide the route length by the number of stops to get the stop spacing

    Args:
      df_route: a dataframe of the stops for a given route
      route_dict: a dictionary containing the route_id, route length in each direction, and the number
    of trips in each direction

    Returns:
      A dictionary with the stop spacing for each direction.
    """
    spc_0 = 0
    if len(df_dir) > 1:
        shape_0 = df_dir.groupby("shape_id").count().trip_id.idxmax()
        n_stops = len(
            df_dir[df_dir.trip_id == df_dir[df_dir.shape_id == shape_0].trip_id.unique()[0]]
        )
        spc_0 = route_dict["route_length"] / n_stops
    return {"stop_spacing": np.round(spc_0, 2) if spc_0 != 0 else 0}


def get_route_lens(df_dir: pd.DataFrame, df_shapes: LineString) -> Dict[str, float]:
    """
    It takes a dataframe of trips for a given route and a dataframe of shapes for a given route, and
    returns a dictionary with the length of the route in each direction

    Args:
      df_route: a dataframe of the routes.txt file
      df_shapes: the shapes dataframe

    Returns:
      A dictionary with the route lengths for each direction.
    """
    epsg_zone = get_zone_epsg_from_ls(df_shapes.iloc[0]["geometry"])
    len_0 = 0
    if len(df_dir) > 1:
        shape_0 = df_dir.groupby("shape_id").count().trip_id.idxmax()
        len_0 = (
            df_shapes.loc[df_shapes.shape_id == shape_0].to_crs(epsg_zone).geometry.length.iloc[0]
        )
    return {"route_length": np.round(len_0 / 1000, 2)}


def get_route_stats(feed: Feed, peak_time: bool = False) -> pd.DataFrame:
    """
    It takes a GTFS feed and a route_id and returns a dataframe with the following columns:

    - route_id
    - route_length
    - route_time
    - average_headway
    - peak_times
    - average_speed
    - min_spacing
    - stop_spacing

    The function is a bit long, but it's not too complicated.

    The first thing it does is merge the stop_times and trips tables. This is necessary because the
    stop_times table doesn't have the route_id column.

    The next thing it does is create an empty dictionary called route_dict. This will be used to store
    the results of the analysis.

    For each route, it creates a new dictionary called ret_dict. This will be used to store the results
    of the analysis for the current

    Args:
      feed: the GTFS feed object
      route_id: The route_id of the route you want to analyze

    Returns:
      A dataframe with the route_id as the first column and the rest columns are the stats for the route.
    """
    df_merge = feed.stop_times.merge(feed.trips, how="left", on="trip_id")
    df_shapes = feed.shapes
    route_list = []
    for route in df_merge.route_id.unique():
        df_route = df_merge[df_merge.route_id == route]
        for direction in df_route.direction_id.unique():
            df_dir = df_route[df_route.direction_id == direction]
            ret_dict = {}
            ret_dict["route"] = route
            ret_dict["direction"] = direction
            ret_dict.update(get_route_lens(df_dir, df_shapes))
            ret_dict.update(get_route_time(df_dir))
            ret_dict.update(get_average_headway(df_dir))
            ret_dict.update(get_average_speed(df_dir, ret_dict))
            ret_dict.update(average_active_buses(df_dir))
            ret_dict.update(get_bus_spacing(ret_dict))
            ret_dict.update(get_stop_spacing(df_dir, ret_dict))
            if peak_time:
                ret_dict.update(get_all_peak_times(df_dir))

            route_list.append(ret_dict)
    df = pd.DataFrame.from_records(route_list)
    return df.reset_index(drop=True)



"""Tests for `gtfs_segments` package."""

import os
import unittest

import requests

import gtfs_segments.partridge_mod as ptg
from gtfs_segments import get_bus_feed

test_dir = os.path.dirname(__file__)


class TestBasic(unittest.TestCase):
    """Tests for some basic prerequisites for the package to work."""

    def setUp(self):
        """Set up test fixtures, if any."""

        self.gtfs_path = os.path.join(
            test_dir,
            "data",
            "Ann Arbor-University of Michigan Transit Services-MI",
            "gtfs.zip",
        )

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_state_abbreviations(self):
        """
        The function tests if a given URL exists and is live.
        """
        url = "https://github.com/UTEL-UIUC/gtfs_segments/blob/main/state_abbreviations.json"

        try:
            response = requests.get(url, timeout=5)
            # Additional check to ensure the URL not only exists but is also live
            self.assertTrue(response.status_code == 200)
        except requests.RequestException as e:
            self.fail(f"URL {url} doesn't exist or couldn't be reached. Error: {e}")

    def test_gtfs_exists(self):
        """
        The function tests if a given GTFS file exists.
        """
        self.assertTrue(
            os.path.exists(self.gtfs_path),
            "The Ann Arbor example gtfs file does not exist",
        )

    def test_partridge(self):
        """
        The function tests if partridge is pointing to the modified version.
        """
        self.assertTrue(
            ptg.__version__ == "1.1.1b1",
            "Please use the following version of partridge 1.1.1b1",
        )
        feed = get_bus_feed(self.gtfs_path)
        self.assertTrue(
            isinstance(feed, ptg.gtfs.Feed),
            "Error with feed type. Make sure the partridge library is installed correctly",
        )
        self.assertGreaterEqual(len(feed.agency), 1, "Some error with feed processing")



"""Tests for `gtfs_segments` package."""

import os
import unittest

import geopandas as gpd

from gtfs_segments.gtfs_segments import get_gtfs_segments, inspect_feed
from gtfs_segments.partridge_func import get_bus_feed

test_dir = os.path.dirname(__file__)


class TestGTFSSegments(unittest.TestCase):
    """Tests for gtfs_segments.py module in the package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.gtfs_path = os.path.join(
            test_dir,
            "data",
            "Ann Arbor-University of Michigan Transit Services-MI",
            "gtfs.zip",
        )

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_feed_inspect(self):
        """
        The function `test_feed_inspect` tests the `inspect_feed` function on a specific feed.
        """
        feed = get_bus_feed(self.gtfs_path)
        self.assertTrue(
            inspect_feed(feed),
            "Error with feed inspection. Should work for the Ann Arbor example feed",
        )

    def test_get_gtfs_segments(self):
        """
        The function `test_get_gtfs_segments` tests various parameters of the `get_gtfs_segments`
        function.
        """
        df = get_gtfs_segments(self.gtfs_path)
        self.assertTrue(
            type(df) == gpd.GeoDataFrame,
            "Error with get_gtfs_segments. Should work for the Ann Arbor example feed",
        )
        df_threshold = get_gtfs_segments(self.gtfs_path, threshold=5)
        self.assertTrue(
            len(df_threshold) <= len(df),
            "Higher threshold should result in fewer segments",
        )
        df_agency = get_gtfs_segments(self.gtfs_path, agency_id="1")
        self.assertTrue(
            len(df_agency) == len(df),
            "Should lead to same number of segments as the original feed has only one agency",
        )
        df_max_spacing = get_gtfs_segments(self.gtfs_path, max_spacing=3000)
        self.assertTrue(
            len(df_max_spacing) <= len(df),
            "Higher max_spacing should result in fewer segments",
        )
        self.assertTrue(
            df_max_spacing["distance"].max() <= 3000,
            "Max spacing should be less than or equal to 3000",
        )
        self.assertTrue(
            df_max_spacing["distance"].min() >= 0,
            "Min spacing should be greater than or equal to 0",
        )



import concurrent.futures
import os
from threading import RLock
from typing import Any, Dict, Optional, Union

import pandas as pd

from .parsers import transforms_dict
from .utilities import detect_encoding

View = Dict[str, Any]


class Feed(object):
    def __init__(
        self,
        source: Union[str, "Feed"],
        view: Optional[View] = None,
    ):
        self._view: View = {} if view is None else view
        self._cache: Dict[str, pd.DataFrame] = {}
        self._pathmap: Dict[str, str] = {}
        self._delete_after_reading: bool = False
        self._shared_lock = RLock()
        self._locks: Dict[str, RLock] = {}
        self._transforms_dict = transforms_dict()
        if isinstance(source, self.__class__):
            self._read = source.get
        elif isinstance(source, str) and os.path.isdir(source):
            self._read = self._read_csv
            self._bootstrap(source)
        else:
            raise ValueError("Invalid source")

        self.file_list = [
            "agency.txt",
            "calendar.txt",
            "calendar_dates.txt",
            "routes.txt",
            "trips.txt",
            "shapes.txt",
            "stop_times.txt",
            "stops.txt",
            "fare_attributes.txt",
            "fare_rules.txt",
            "feed_info.txt",
            "frequencies.txt",
            "transfers.txt",
        ]

    def __getattr__(self, name: str) -> pd.DataFrame:
        if name in [f[:-4] for f in self.file_list]:
            return self.get(f"{name}.txt")
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def get(self, filename: str) -> pd.DataFrame:
        lock = self._locks.get(filename, self._shared_lock)
        with lock:
            df = self._cache.get(filename)
            if df is None:
                df = self._read(filename)
                df = self._filter(filename, df)
                self._convert_types(filename, df)
                df = df.reset_index(drop=True)
                df = self._transform(filename, df)
                self.set(filename, df)
            return self._cache[filename]

    def set(self, filename: str, df: pd.DataFrame) -> None:
        lock = self._locks.get(filename, self._shared_lock)
        with lock:
            self._cache[filename] = df

    def _bootstrap(self, path: str) -> None:
        # Walk recursively through the directory
        for root, _subdirs, files in os.walk(path):
            for fname in files:
                basename = os.path.basename(fname)
                if basename in self._pathmap:
                    # Verify that the folder does not contain multiple files of the same name.
                    raise ValueError("More than one {} in folder".format(basename))
                # Index paths by their basename.
                self._pathmap[basename] = os.path.join(root, fname)
                # Build a lock for each file to synchronize reads.
                self._locks[basename] = RLock()

    def _read_csv(self, filename: str) -> pd.DataFrame:
        path = self._pathmap.get(filename)

        if path is None or os.path.getsize(path) == 0:
            # The file is missing or empty. Return an empty
            # DataFrame containing any required columns.
            return pd.DataFrame()

        # If the file isn't in the zip, return an empty DataFrame.
        with open(path, "rb") as f:
            encoding = detect_encoding(f)
        # file_columns = self._transforms_dict[filename].get("usecols", [])
        # df = pd.read_csv(path, usecols=file_columns, dtype=str, encoding=encoding, index_col=False)
        df_head = pd.read_csv(
            path, header=0, dtype=str, encoding=encoding, engine="c", index_col=False, nrows=1
        )
        available_columns = set(df_head.columns)
        file_columns = self._transforms_dict[filename].get("usecols", [])
        if len(file_columns) != 0:
            use_cols = list(set(file_columns.keys()).intersection(available_columns))
            df = pd.read_csv(
                path,
                usecols=use_cols,
                header=0,
                dtype="str",
                encoding=encoding,
                engine="c",
                index_col=False,
            )
        else:
            df = pd.read_csv(
                path,
                header=0,
                dtype="str",
                encoding=encoding,
                engine="c",
                index_col=False,
            )
        # Strip leading/trailing whitespace from column names
        # df.rename(columns=lambda x: x.strip(), inplace=True)

        # if not df.empty:
        #     # Strip leading/trailing whitespace from column values
        #     for col in df.columns:
        #         df.loc[:, col] = df[col].str.strip()

        return df

    def _filter(self, filename: str, df: pd.DataFrame) -> pd.DataFrame:
        """Apply view filters"""
        view = self._view.get(filename)
        if view is None:
            return df

        for col, values in view.items():
            # If applicable, filter this dataframe by the given set of values
            if col in df.columns:
                df = df[df[col].isin(set(values))]
        return df

    def _convert_types(self, filename: str, df: pd.DataFrame) -> None:
        """
        Apply type conversions
        """
        if df.empty:
            return

        converters = self._transforms_dict[filename].get("converters", {})
        for col, converter in converters.items():
            if col in df.columns:
                df.loc[:, col] = converter(df.loc[:, col])

    def _transform(self, filename: str, df: pd.DataFrame) -> pd.DataFrame:
        transformations = self._transforms_dict[filename].get("transformations", [])
        if "geometry" not in df.columns:
            for transform in transformations:
                df = transform(df)

        return df


def fetch_data(feed: Feed, property_name: str) -> None:
    """Function to fetch data for a given property name."""
    getattr(feed, property_name)


def parallel_read(feed: Feed) -> None:
    """
    Fetches GTFS data in parallel using ThreadPoolExecutor.

    Args:
        feed (Feed): The GTFS feed object.

    Returns:
        None
    """
    property_names = [
        "stop_times",
        "stops",
        "agency",
        "calendar",
        "calendar_dates",
        "routes",
        "trips",
        "shapes",
    ]

    # Use ThreadPoolExecutor to fetch data in parallel
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        # Create a future for each property
        {executor.submit(fetch_data, feed, name): name for name in property_names}



"""Tests for `gtfs_segments` package."""

import glob
import os
import unittest

import pandas as pd

from gtfs_segments import download_latest_data, fetch_gtfs_source

test_dir = os.path.dirname(__file__)
test_folder = os.path.join(test_dir, "output")
if not os.path.exists(test_folder):
    os.mkdir(test_folder)


class TestMobiliy(unittest.TestCase):
    """Tests for the mobility.py module."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.gtfs_folder = os.path.join(
            test_dir,
            "data",
            "San Francisco-San Francisco Municipal Transportation Agency (SFMTA, Muni)-CA",
        )
        self.gtfs_path = os.path.join(
            test_dir,
            "data",
            "San Francisco-San Francisco Municipal Transportation Agency (SFMTA, Muni)-CA",
            "gtfs.zip",
        )

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_check_fetch_all(self):
        """
        The function "check_fetch_all" checks if the output of the "fetch_gtfs_source" function is a
        dataframe and if the number of all feeds is greater than or equal to the number of active feeds.
        """
        df_all_active = fetch_gtfs_source("ALL", active=True)
        df_all = fetch_gtfs_source("ALL", active=False)

        self.assertEqual(
            type(df_all_active),
            pd.DataFrame,
            "The output of fetch_gtfs_source should be a dataframe",
        )
        self.assertGreaterEqual(
            len(df_all),
            len(df_all_active),
            "The number of all feeds should be greater than or equal to the number of active feeds",
        )

    def test_download_gtfs(self):
        sources_df = fetch_gtfs_source("SFMTA", active=True)
        download_latest_data(sources_df, os.path.join(test_dir, "data"))
        self.assertTrue(
            os.path.exists(os.path.join(test_dir, "data")),
            "Check if the data_test folder exists",
        )
        folders = glob.glob(os.path.join(test_dir, "data/*"))
        self.assertTrue(len(folders) > 0, "Check if the data_test folder is not empty")
        sub_folder = glob.glob(os.path.join(test_dir, "data/*"))
        file = os.listdir(sub_folder[0])[0]
        self.assertTrue(file.endswith(".zip"), "Check if the data folder contains a zip file")
        if os.path.exists(self.gtfs_path):
            os.remove(self.gtfs_path)
            # You can remove and empty folders with os.rmdir()
            os.rmdir(self.gtfs_folder)



from setuptools import find_packages, setup

__version__ = "0.2.0"
DESCRIPTION = "GTFS segments"
with open("README.md", "r", encoding="utf8") as fh:
    LONG_DESCRIPTION = fh.read()

REQUIREMENTS = [
    "geopandas",
    "scipy",
    "shapely",
    "numpy",
    "pandas",
    "matplotlib",
    "utm",
    "contextily",
    "requests",
    "isoweek",
    "faust-cchardet",
    "charset_normalizer",
    "folium",
    "thefuzz",
]

# Setting up
setup(
    name="gtfs_segments",
    version=__version__,
    author="Saipraneeth Devunuri",
    author_email="<sd37@illinois.edu>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    url="https://github.com/UTEL-UIUC/gtfs_segments",
    keywords=["python", "gtfs", "geodata"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "License :: OSI Approved :: MIT License",
    ],
)
